#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cmath>

#define KERNEL_SIZE 3

// GELU activation function
__device__ inline float gelu_activate(float x) {
    return 0.5f * x * (1.f + erff(x / 1.41421356f));
}

// Warp-level reduction using shuffle intrinsics
__device__ inline float warp_reduce_sum(float val) {
    // Assuming warp size is 32
    for (int offset = 16; offset > 0; offset /= 2)
        val += __shfl_down_sync(0xffffffff, val, offset);
    return val;
}

// Compute convolution for a fixed 3x3 kernel using unrolling and shared weight
__device__ inline float compute_conv_unrolled_shared(
    const float* __restrict__ input,
    const float* __restrict__ sweight,
    const int n,
    const int in_channels,
    const int in_h,
    const int in_w,
    const int row,
    const int col
) {
    float sum = 0.0f;
    // Loop over input channels with unrolling for a 3x3 kernel
    #pragma unroll
    for (int c = 0; c < in_channels; c++) {
        int input_channel_offset = n * in_channels * in_h * in_w + c * in_h * in_w;
        int weight_channel_offset = c * (KERNEL_SIZE * KERNEL_SIZE);
        // Unrolled 3x3 convolution
        sum += __ldg(&input[input_channel_offset + (row + 0) * in_w + (col + 0)]) * sweight[weight_channel_offset + 0];
        sum += __ldg(&input[input_channel_offset + (row + 0) * in_w + (col + 1)]) * sweight[weight_channel_offset + 1];
        sum += __ldg(&input[input_channel_offset + (row + 0) * in_w + (col + 2)]) * sweight[weight_channel_offset + 2];

        sum += __ldg(&input[input_channel_offset + (row + 1) * in_w + (col + 0)]) * sweight[weight_channel_offset + 3];
        sum += __ldg(&input[input_channel_offset + (row + 1) * in_w + (col + 1)]) * sweight[weight_channel_offset + 4];
        sum += __ldg(&input[input_channel_offset + (row + 1) * in_w + (col + 2)]) * sweight[weight_channel_offset + 5];

        sum += __ldg(&input[input_channel_offset + (row + 2) * in_w + (col + 0)]) * sweight[weight_channel_offset + 6];
        sum += __ldg(&input[input_channel_offset + (row + 2) * in_w + (col + 1)]) * sweight[weight_channel_offset + 7];
        sum += __ldg(&input[input_channel_offset + (row + 2) * in_w + (col + 2)]) * sweight[weight_channel_offset + 8];
    }
    return sum;
}

// Fused kernel: convolution, bias addition, GELU activation, and global average pooling
// Each block computes one (n, c_out) pair over the convolution output spatial dimensions (out_h x out_w).
extern "C" __global__ void fused_conv_gelu_pool_kernel(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    const int N,
    const int in_channels,
    const int in_h,
    const int in_w,
    const int out_channels,
    const int out_h,
    const int out_w
) {
    // Block assignment: blockIdx.y -> batch index (n), blockIdx.x -> output channel (c_out)
    const int n = blockIdx.y;
    const int c_out = blockIdx.x;
    const int total_pixels = out_h * out_w;

    // Shared memory for loading the filter weights for the current output channel.
    // Size: in_channels * KERNEL_SIZE * KERNEL_SIZE floats.
    extern __shared__ float sweight[];
    const int numWeights = in_channels * KERNEL_SIZE * KERNEL_SIZE;

    // Load weights from global memory into shared memory once per block
    for (int i = threadIdx.x; i < numWeights; i += blockDim.x) {
        sweight[i] = weight[c_out * numWeights + i];
    }
    __syncthreads();

    float local_sum = 0.0f;

    // Each thread processes multiple output pixels using a grid-stride loop
    for (int idx = threadIdx.x; idx < total_pixels; idx += blockDim.x) {
        int row = idx / out_w;
        int col = idx % out_w;
        float conv_val = compute_conv_unrolled_shared(input, sweight, n, in_channels, in_h, in_w, row, col);
        conv_val += bias[c_out];
        conv_val = gelu_activate(conv_val);
        local_sum += conv_val;
    }

    // Warp-level reduction
    local_sum = warp_reduce_sum(local_sum);

    // Shared memory for block-level reduction among warps
    __shared__ float shared_sum[32]; // Supports up to 32 warps per block
    int lane = threadIdx.x & 31;
    int warpId = threadIdx.x >> 5;
    if (lane == 0) {
        shared_sum[warpId] = local_sum;
    }
    __syncthreads();

    float block_sum = 0.0f;
    int numWarps = (blockDim.x + 31) >> 5;
    if (threadIdx.x < numWarps) {
        block_sum = shared_sum[threadIdx.x];
    }
    block_sum = warp_reduce_sum(block_sum);

    // Thread 0 writes the final averaged result to global memory
    if (threadIdx.x == 0) {
        output[n * out_channels + c_out] = block_sum / float(total_pixels);
    }
}

// Host function that launches the fused kernel
torch::Tensor forward(
    torch::Tensor input,
    torch::Tensor conv_weight,
    torch::Tensor conv_bias
) {
    TORCH_CHECK(input.is_cuda(), "input must be a CUDA tensor");
    TORCH_CHECK(conv_weight.is_cuda(), "conv_weight must be a CUDA tensor");
    TORCH_CHECK(conv_bias.is_cuda(), "conv_bias must be a CUDA tensor");

    const int N = input.size(0);
    const int in_channels = input.size(1);
    const int in_h = input.size(2);
    const int in_w = input.size(3);
    const int out_channels = conv_weight.size(0);
    // For a 3x3 filter, output dimensions are (in_h - 2, in_w - 2)
    const int out_h = in_h - 2;
    const int out_w = in_w - 2;

    auto options = torch::TensorOptions().dtype(input.dtype()).device(input.device());
    // Final output has shape (N, out_channels) after global average pooling
    auto final_output = torch::empty({N, out_channels}, options);

    // Launch the fused kernel with a 2D grid: (out_channels, N)
    dim3 grid(out_channels, N);
    const int threads = 256;
    // Shared memory size: in_channels * 3 * 3 * sizeof(float)
    size_t shared_mem = in_channels * KERNEL_SIZE * KERNEL_SIZE * sizeof(float);

    fused_conv_gelu_pool_kernel<<<grid, threads, shared_mem>>>(
        input.data_ptr<float>(),
        conv_weight.data_ptr<float>(),
        conv_bias.data_ptr<float>(),
        final_output.data_ptr<float>(),
        N, in_channels, in_h, in_w,
        out_channels, out_h, out_w
    );

    return final_output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused Conv2d + GELU + GlobalAvgPool with Shared Weight Tiling");
}
